/*
 * High-level commands for wheel control.
 *
 * Author: BLADOWSKI Jakub
 * S4 Project 63: Follower Robot
 * Telecom Bretagne
 * February - June 2015
 */
 
#include "movement.h"

/*
 * Speed of each wheel (in turns per second), flag signaling ongoing movement and time instant at which movement should stop.
 * Separately for two kinds of movement: displacement and rotation.
 */
float dis_tps[3];
boolean displacement = false;
//float goal[3];
unsigned long dis_stop_time;
float rot_tps[3];
boolean rotation = false;
unsigned long rot_stop_time;

/*
 * Helper function for converting angle values.
 */
float degToRad(float angle_deg) {
    return 3.1416 * angle_deg / 180.;
}

/*
 * Set desired motor speeds following a displacement command.
 */
void command_displacement(float distance_m, float angle) {
  if (!rotation) {
    /*goal[0] = distance_m * sin(degToRad(angle));
    goal[1] = distance_m * sin(degToRad(angle - 120));
    goal[2] = distance_m * sin(degToRad(angle - 240));*/
    //ctr = 0;
    float max_speed = 1000;
    dis_tps[0] = max_speed * sin(degToRad(angle));
    dis_tps[1] = max_speed * sin(degToRad(angle - 120));
    dis_tps[2] = max_speed * sin(degToRad(angle - 240));
    
    // coefficient found experimentally for 1 meter
    dis_stop_time = millis() + 3e6 * distance_m / max_speed;
    displacement = true;
  }
}

/*
 * Helper for finding coeffitient for rotation time given angle.
 * Coeffitient value found experimentally at selected angles, interpolated between.
 */
float rotation_coeff(float angle) {
  float angles[] = {0, 8, 15, 30, 45, 60 ,90, 120, 180};
  float coeffs[] = {0, 320e3, 405e3, 465e3, 585e3, 660e3, 900e3, 1080e3, 1400e3};
  int i = 0;
  while (true) {
    if (i == 8) return coeffs[8];
    if (angle < angles[i+1]) break;
    i += 1;
  }
  return (angle-angles[i])/(angles[i+1]-angles[i])*(coeffs[i+1]-coeffs[i]) + coeffs[i];
}

/*
 * Set desired motor speeds following a rotation command.
 */
void command_rotation(float angle) {
  if (!rotation) {
    float rotation_speed = 500;
  
    if (angle < 0) {
      rotation_speed = -rotation_speed;
      angle = -angle;
    }
    
    rot_tps[0] = rotation_speed;
    rot_tps[1] = rotation_speed;
    rot_tps[2] = rotation_speed;
    
    rot_stop_time = millis() + rotation_coeff(angle) / abs(rotation_speed);
    rotation = true;
  }
}

/*
 * Set desired motor speeds to zero following a halt command.
 */
void command_halt() {
  rotation = false;
  displacement = false;
  for(int i = 0; i < 3; i++) {
    dis_tps[i] = 0;
    rot_tps[i] = 0;
  }
}

/*
 * Check condition for stopping wheel movement due to displacement.
 */
bool condition_stop_displacement() {
  return displacement && (millis() >= dis_stop_time);
}

/*
 * Check condition for stopping wheel movement due to rotation.
 */
bool condition_stop_rotation() {
  return rotation && (millis() >= rot_stop_time);
}

/*
 * Set displacement component of motor speeds to zero at scheduled stop of displacement.
 */
void stop_displacement() {
  displacement = false;
  for(int i = 0; i < 3; i++)
    dis_tps[i] = 0;
}

/*
 * Set rotation component of motor speeds to zero at scheduled stop of rotation.
 */
void stop_rotation() {
  rotation = false;
  for(int i = 0; i < 3; i++)
    rot_tps[i] = 0;
}

/*
 * Return the total desired speed in turns per second of the given motor.
 */
float desired_tps(int motor) {
  return dis_tps[motor] + rot_tps[motor];
}
