/*
 * PID control of wheel speed.
 * Controller output: duty cycle of PWM signal controling wheel speed.
 * Error signal: difference between desired and measured wheel speed in turns per PID time step.
 *
 * Author: BLADOWSKI Jakub
 * S4 Project 63: Follower Robot
 * Telecom Bretagne
 * February - June 2015
 */
 
#include "pid.h"

/*
 * Length of PID time step in microseconds.
 */
#define PID_PERIOD 5e4

/*
 * PID coefficients.
 */
#define KP 1e-4
#define KI 0
#define KD 2e-5

/*
 * Instant at which current time step ends; controller output; integral of previous error values; last error value.
 */
unsigned long pid_next_tic;
float pwm_value[3];
float integral[3];
float last_error[3];

/*
 * Initialize variables.
 */
void setup_pid() {
  pid_next_tic = micros() + PID_PERIOD;
  reset_encoding();
  for(int i = 0; i < 3; i++) {
    pwm_value[i] = 0;
    integral[i] = 0;
    last_error[i] = 0;
  }
}

/*
 * Check for end of time step.
 */
bool condition_pid_tic() {
  return micros() >= pid_next_tic;
}

/*
 * End time step.
 */
void pid_tic() {
  pid_next_tic = micros() + PID_PERIOD;
  pid_control(MOTOR_1);
  pid_control(MOTOR_2);
  pid_control(MOTOR_3);
  reset_encoding();
}

/*
 * Calculate controller output for given motor at the end of a time step.
 */
void pid_control(int motor) {
  float expected = desired_tps(motor) * (PID_PERIOD/1e6);
  float error = expected - get_encoding(motor);
  integral[motor] += error * (PID_PERIOD/1e6);
  float derivative = (error - last_error[motor]) / (PID_PERIOD/1e6);
  
  pwm_value[motor] += KP * error + KI * integral[motor] + KD * derivative;
  pwm_value[motor] = constrain(pwm_value[motor], -1, 1);
  if (desired_tps(motor) == 0)
    if (abs(get_encoding(motor)) < 3) {
      pwm_value[motor] = 0;
      integral[motor] = 0;
    }
  set_motor_speed(motor, pwm_value[motor]);
  last_error[motor] = error;
}

/*
 * Disable PID control by making it impossible to reach the next tic.
 */
void disable_pid() {
  pid_next_tic = 4294967295;
}
